from dataclasses import dataclass, field
from click import MissingParameter
import urllib3
from os import path
from sys import path as syspath
from sys import exit as sysexit
from os import makedirs 
import json

@dataclass
class Config:
    """
    Dataclass for huetui config.
    """

    tui_settings: dict = field(default_factory=dict)
    ip: str = ""
    api_user: str = ""


class Util:
    """
    Metaclass containing some utility methods
    """

    @staticmethod
    def value_to_percent(value, whole) -> int:
        """Converts a value to a percentage

        Args:
            value (number): value to convert
            whole (number): value of entire scale

        Returns:
            int: percentage
        """
        return round(((value / whole) * 100))

    @staticmethod
    def percent_to_value(percent, whole) -> int:
        """Converts a percentage to a value

        Args:
            percent (number): percentage to convert
            whole (number): value of entire scale

        Returns:
            int: value
        """
        return round((percent / 100) * whole)

    @staticmethod
    def api_get(address: str) -> dict:
        """Make a get request to the given address.

        Args:
            address (string): The address to make the request to. Defaults to None.

        Returns:
            dict: response from api.
        """
        if address:
            # try to create a connection
            http = urllib3.PoolManager()
            response = http.request("GET", address)

            # check if response ok
            if (response.status == 200) and not (
                response.data.decode("utf-8").startswith('[{"error":')
            ):
                payload = json.loads(response.data.decode("utf-8"))
                return dict(payload)

            else:
                raise Exception(
                    "Code: "
                    + str(response.status)
                    + "; API error: "
                    + response.data.decode("utf-8")
                )

        else:
            raise MissingParameter("adress")

    @staticmethod
    def api_put(address: str, data: str) -> bool:
        """Make a put request to the given address.

        Args:
            address (string, optional): The address to make the request to. Defaults to None.
            data (string, optional): The data to put. Defaults to None.

        Returns:
            bool: post successfull.
        """
        if address and data:
            http = urllib3.PoolManager()
            response = http.request(
                "PUT", address, body=data, headers={"Content-Type": "application/json"}
            )

            if (response.status == 200) and (
                response.data.decode("utf-8").startswith('[{"success":')
            ):
                return True
            else:
                raise Exception(
                    "Code: "
                    + str(response.status)
                    + "; API error: "
                    + response.data.decode("utf-8")
                )

        else:
            raise MissingParameter("adress or data")

    @staticmethod
    def __ensure_dir(paf: str) -> bool:
        dir = path.dirname(paf)
        if not path.exists(dir):
            makedirs(dir)
            return False
        else:
            return True
            
    @staticmethod
    def generate_config_file(ip: str, api_user: str, config_path: str) -> None:
        """Generate a config file with default values.
        """
        config = path.expanduser(config_path)
        Util._Util__ensure_dir(config)

        with open(config, 'w', encoding='utf-8') as f:
            f.write(
                "# sample config. Edit the dict to your liking.\n"
                "import py_cui.colors as colors\n"
                "from huetui.backend.utils import Config\n"
                "c = Config()\n"
                f"c.ip = \"{ip}\"\n"
                f"c.api_user = \"{api_user}\"\n"
                "c.tui_settings = {\n"
                "\t\"unicode\": True,\n"
                "\t\"wallpaper\": \"~/Path/To/Wallpaper.jpg\",\n"
                "\t\"uicolors\": {\n"
                "\t\t\"main_ui\": colors.WHITE_ON_BLACK,\n"
                "\t\t\"selected\": colors.CYAN_ON_BLACK,\n"
                "\t\t\"border\": colors.BLUE_ON_BLACK,\n"
                "\t\t\"logo\": colors.CYAN_ON_BLACK,\n"
                "\t\t\"status_bar\": colors.BLACK_ON_WHITE,\n"
                "\t\t\"title_bar\": colors.BLACK_ON_MAGENTA,\n"
                "\t},\n"
                "}\n"
            )
    
    @staticmethod
    def load_config_file(config_path: str) -> Config:
        """Load a config file and return config dataclass.

        Args:
            config_path (str): path to config file.

        Returns:
            Config: config object.
        """
        config = path.expanduser(config_path)
        syspath.append(path.dirname(path.expanduser(config_path)))
        try:
            from config import c
            return c
        except ModuleNotFoundError:
            print("Config file not found. Generating a new config file.")
            ip = input("Bridge ip: ")
            api_user = input("Bridge api user: ")
            Util.generate_config_file(ip, api_user, config_path)
            print("Config file generated. Rerun huetui.")
            sysexit(0)


import py_cui as cui
from py_cui import keys as Keys

from huetui.backend.bridge import Bridge
from huetui.backend.utils import Config

from huetui.frontend.lights_menu import LightMenu
from huetui.frontend.groups_menu import GroupMenu
from huetui.frontend.scenes_menu import SceneMenu

import time, threading


class extPyCUI(cui.PyCUI):
    """Extends the PyCUI class to add a light menu widget.

    Args:
        x (int): width of the window
        y (int): height of the window
        bridge (Bridge): bridge object
    """

    def __init__(self, x: int, y: int, bridge: Bridge) -> None:
        self.bridge = bridge
        super(extPyCUI, self).__init__(x, y)

    def add_light_menu(
        self,
        title: str,
        row: int,
        column: int,
        row_span: int,
        column_span: int,
        padx=1,
        pady=1,
    ) -> None:
        """Adds a light menu widget to the root.

        Args:
            title (str): menu title
            row (int): x position of the menu
            column (int): y position of the menu
            row_span (int): x width of the menu
            column_span (int): y height of the menu
            padx (int, optional): x padding .Defaults to 0.
            pady (int, optional): y padding. Defaults to 0.

        """
        id = f"Widget{len(self._widgets.keys())}"
        new_light_menu = LightMenu(
            id,
            title,
            self._grid,
            row,
            column,
            row_span,
            column_span,
            padx,
            pady,
            self._logger,
            self.bridge,
            self,
        )
        self._widgets[id] = new_light_menu
        if self.set_selected_widget is None:
            self.set_selected_widget(id)
        return new_light_menu

    def add_group_menu(
        self,
        title: str,
        row: int,
        column: int,
        row_span: int,
        column_span: int,
        padx=1,
        pady=1,
    ) -> None:
        """Adds a group menu widget to the root.

        Args:
            title (str): menu title
            row (int): x position of the menu
            column (int): y position of the menu
            row_span (int): x width of the menu
            column_span (int): y height of the menu
            padx (int, optional): x padding .Defaults to 0.
            pady (int, optional): y padding. Defaults to 0.

        """
        id = f"Widget{len(self._widgets.keys())}"
        new_group_menu = GroupMenu(
            id,
            title,
            self._grid,
            row,
            column,
            row_span,
            column_span,
            padx,
            pady,
            self._logger,
            self.bridge,
            self,
        )
        self._widgets[id] = new_group_menu
        if self.set_selected_widget is None:
            self.set_selected_widget(id)
        return new_group_menu

    def add_scene_menu(
        self,
        title: str,
        row: int,
        column: int,
        row_span: int,
        column_span: int,
        padx=1,
        pady=1,
    ) -> None:
        """Adds a scene menu widget to the root.

        Args:
            title (str): menu title
            row (int): x position of the menu
            column (int): y position of the menu
            row_span (int): x width of the menu
            column_span (int): y height of the menu
            padx (int, optional): x padding .Defaults to 0.
            pady (int, optional): y padding. Defaults to 0.

        """
        id = f"Widget{len(self._widgets.keys())}"
        new_scene_menu = SceneMenu(
            id,
            title,
            self._grid,
            row,
            column,
            row_span,
            column_span,
            padx,
            pady,
            self._logger,
            self.bridge,
            self,
        )
        self._widgets[id] = new_scene_menu
        if self.set_selected_widget is None:
            self.set_selected_widget(id)
        return new_scene_menu


class MainWindow:
    """Main HueTUI window class."""

    def __init__(self, master: extPyCUI, bridge: Bridge, config: Config) -> None:
        self.master = master
        self.bridge = bridge
        self._update_thread = None
        self.config = config

        # add menus to root
        self.logo = self.master.add_block_label(self._get_logo_text(), 0, 0, 1, 2)
        self.logo.set_selectable(False)
        self.info_menu = self.master.add_text_block(
            "Bridge", 0, 2, 1, 1, initial_text=self.bridge.info_as_str()
        )
        self.info_menu.set_selectable(False)
        self.key_menu = self.master.add_text_block(
            "Usage",
            0,
            3,
            1,
            1,
            initial_text="Movement: arrow-keys\nSelect: enter\nExit: esc\nQuit: q\nWallpaper colors: w",
        )
        self.key_menu.set_selectable(False)
        self.light_menu = self.master.add_light_menu("Lights", 1, 0, 2, 2)
        self.group_menu = self.master.add_group_menu("Groups", 3, 0, 2, 2)
        self.scene_menu = self.master.add_scene_menu("Scenes", 1, 2, 2, 2)
        self.active_menu = self.master.add_scroll_menu("Active", 3, 2, 2, 2, 1, 1)
        self.active_menu.set_selectable(False)

        # add keybindings
        self.master.add_key_command(Keys.KEY_W_LOWER, self._set_wallpaper_colors)

        # hook stop function to exit
        self.master.run_on_exit(self._stop_active_devices_thread)

        # refresh every 10ms
        self.master.set_refresh_timeout(0.1)

        # toggle unicode
        if config.tui_settings["unicode"]:
            self.master.toggle_unicode_borders()

        # start update thread
        self._init_active_devices_thread()

        # set ui colors
        for id in self.master._widgets.keys():
            self.master._widgets[id].set_color(
                config.tui_settings["uicolors"]["main_ui"]
            )
            self.master._widgets[id].set_selected_color(
                config.tui_settings["uicolors"]["selected"]
            )
            self.master._widgets[id].set_border_color(
                config.tui_settings["uicolors"]["border"]
            )
            self.master.status_bar.set_color(
                config.tui_settings["uicolors"]["status_bar"]
            )
            self.master.title_bar.set_color(
                config.tui_settings["uicolors"]["title_bar"]
            )
            self.logo.set_color(config.tui_settings["uicolors"]["logo"])

        self.active_menu.set_selected_color(config.tui_settings["uicolors"]["main_ui"])

    def _update_active_menu(self) -> None:
        """update the active menu with the active devices"""
        t = threading.currentThread()

        # should de thread be running?
        while getattr(t, "do_run", True):
            # list for all active devices
            active = []

            for light in self.bridge.lights:
                if light.on:
                    # add light to active list
                    format_str = f"({light.brightness}%) {light.name}"
                    active.append(format_str)

            # refresh active menu
            self.active_menu.clear()
            self.active_menu.add_item_list(active)

            time.sleep(0.1)

    def _init_active_devices_thread(self) -> None:
        """initializes the thread that updates the active devices menu"""
        # create new thread and start it
        self._update_thread = threading.Thread(target=self._update_active_menu)
        self._update_thread.start()

    def _stop_active_devices_thread(self) -> None:
        """stops the active devices update thread"""
        # stop the thread
        self._update_thread.do_run = False

    def _get_logo_text(self) -> str:
        """returns the logo banner with linebreaks"""

        logo = "                                                            \n"
        logo += "██╗  ██╗██╗   ██╗███████╗              ████████╗██╗   ██╗██╗\n"
        logo += "██║  ██║██║   ██║██╔════╝              ╚══██╔══╝██║   ██║██║\n"
        logo += "███████║██║   ██║█████╗      █████╗       ██║   ██║   ██║██║\n"
        logo += "██╔══██║██║   ██║██╔══╝      ╚════╝       ██║   ██║   ██║██║\n"
        logo += "██║  ██║╚██████╔╝███████╗                 ██║   ╚██████╔╝██║\n"
        logo += "╚═╝  ╚═╝ ╚═════╝ ╚══════╝                 ╚═╝    ╚═════╝ ╚═╝\n"

        return logo

    def _set_wallpaper_colors(self) -> None:
        """wrapper function to set the wallpaper colors and showing a status message"""
        self.master.show_loading_icon_popup(
            "Wallpaper colors", "Extracting and setting colors"
        )
        thread = threading.Thread(target=self._thread_set_wallpaper_colors)
        thread.start()

    def _thread_set_wallpaper_colors(self):
        """sets the lights color to the current colors of the wallpaper"""
        try:
            self.bridge.set_lights_from_image(self.config.tui_settings["wallpaper"])
            self.master.stop_loading_popup()
        except FileNotFoundError or KeyError:
            self.master.stop_loading_popup()
            self.master.show_error_popup(
                "Error",
                "Wallpaper "
                + self.config.tui_settings["wallpaper"]
                + " not found. Enter to continue.",
            )



import py_cui.keys as Keys
import py_cui.grid as Grid
import py_cui.widgets as Widgets

from py_cui.debug import PyCUILogger as Logger
from huetui.backend.bridge import Bridge
from huetui.backend.light import RGB
from huetui.frontend.utils import get_xrdb_colors


class GroupMenu(Widgets.ScrollMenu):
    """Group menu widget.

    Args:
        title (str): menu title
        grid (Grid): grid object
        row (int): x position of the menu
        column (int): y position of the menu
        row_span (int): x width of the menu
        column_span (int): y height of the menu
        padx (int): x padding
        pady (int): y padding
        logger (Logger): logger object
        bridge (Bridge): bridge object
    """

    def __init__(
        self,
        id,
        title: str,
        grid: Grid,
        row: int,
        column: int,
        row_span: int,
        column_span: int,
        padx: int,
        pady: int,
        logger: Logger,
        bridge: Bridge,
        master,
    ) -> None:
        # set fallback colors in case xrdb is not available
        self.colors = {
            "black": "#000000",
            "gray": "#808080",
            "red": "#800000",
            "light red": "#ff0000",
            "green": "#008000",
            "light green": "#00ff00",
            "yellow": "#808000",
            "light yellow": "#ffff00",
            "blue": "#000080",
            "light blue": "#0000ff",
            "magenta": "#800080",
            "light magenta": "#ff00ff",
            "cyan": "#008080",
            "light cyan": "#00ffff",
            "white": "#c0c0c0",
            "light white": "#ffffff",
        }
        # try get colors from xrdb
        xrdb_colors = get_xrdb_colors()
        if xrdb_colors is not None:
            self.colors = xrdb_colors

        self.master = master
        self.groups = bridge.groups
        super(GroupMenu, self).__init__(
            id, title, grid, row, column, row_span, column_span, padx, pady, logger
        )
        self.set_help_text(
            "Groups: arrow keys to navigate,"
            " j and k to change brightness,"
            " c to set color, Enter to toggle,"
            " ESC to exit"
        )
        self.add_item_list([group.name for group in self.groups])
        self.add_key_command(Keys.KEY_ENTER, self.toggle)
        self.add_key_command(Keys.KEY_K_LOWER, command=self.inc_bri)
        self.add_key_command(Keys.KEY_J_LOWER, command=self.dec_bri)
        self.add_key_command(Keys.KEY_C_LOWER, self.popup)

    def toggle(self) -> None:
        """Toggle group on/off."""
        self.groups[self.get_selected_item_index()].all_on = not self.groups[
            self.get_selected_item_index()
        ].all_on

    def inc_bri(self) -> None:
        """Increase brightness of group."""
        self.groups[self.get_selected_item_index()].brightness += 10

    def dec_bri(self) -> None:
        """Decrease brightness of group."""
        self.groups[self.get_selected_item_index()].brightness -= 10

    def popup(self) -> None:
        """Show color picker popup."""
        self.master.show_menu_popup("Color", self.colors.keys(), self.set_picked_color)

    def set_picked_color(self, color: str) -> None:
        """set picked color by converting it to rgb.

        Args:
            color (str): color name
        """
        selCol = self.colors[color].strip("#")
        rgbTup = tuple(int(selCol[i : i + 2], 16) for i in (0, 2, 4))
        rgbCol = RGB(rgbTup[0], rgbTup[1], rgbTup[2])
        self.groups[self.get_selected_item_index()].on = True
        for light in self.groups[self.get_selected_item_index()].lights:
            light.color = rgbCol



from huetui.frontend.main_window import MainWindow, extPyCUI
from huetui.backend.bridge import Bridge
from huetui.backend.utils import Config


class Root:
    """Root class for the frontend."""
    def __init__(
        self, x: int, y: int, title: str, bridge: Bridge, config: Config
    ) -> None:
        root = extPyCUI(x, y, bridge)
        root.set_title(title)
        MainWindow(root, bridge, config)
        root.start()
        pass


import sys
import subprocess


def get_xrdb_colors() -> dict[str, str] | None:
    """
    Returns:
        None: if no xrdb is present on system
        dict[str, str]: a dict of color names and corresponding hex values
    """
    colors = bytearray()
    err = None
    if sys.platform == "linux":
        try:
            colors, err = subprocess.Popen(
                ["xrdb", "-query"], stdout=subprocess.PIPE
            ).communicate()
        # parse colors
        except Exception:
            pass
        finally:
            if err is None and len(colors) > 0:
                # xrdb is working
                colors = colors.decode().split("\n")
                colors = [
                    col.split("\t")[1] for col in colors if col.startswith("*.color")
                ]
                return {
                    "black": colors[0],
                    "gray": colors[1],
                    "red": colors[2],
                    "light red": colors[3],
                    "green": colors[4],
                    "light green": colors[5],
                    "yellow": colors[6],
                    "light yellow": colors[7],
                    "blue": colors[8],
                    "light blue": colors[9],
                    "magenta": colors[10],
                    "light magenta": colors[11],
                    "cyan": colors[12],
                    "light cyan": colors[13],
                    "white": colors[14],
                    "light white": colors[15],
                }
    return None



from huetui.backend.utils import Util
import huetui.backend.bridge as Bridge


class Group(Util):
    """Class representing a group of lights.

    Args:
        Util (class): parent class containing some utility methods

    Attributes:
        _bridge (class): bridge the group is defined on
        _gid (int): id of the group
        _name (str): name of the group
        _lights (list): list of lights in the group
        _all_on (bool): on/off state of the group
    """

    def __init__(self, bridge: Bridge, gid: int, name: str, lights: list) -> None:
        self._bridge = bridge
        self._gid = gid
        self._name = name
        self._lights = lights
        self._all_on = None
        self._brightness = None

    def __str__(self) -> str:
        return "<class 'Group'> gid: {}".format(self._gid)

    def __repr__(self) -> str:
        return "<class 'Group'> gid: {}".format(self._gid)

    @property
    def gid(self) -> int:
        return self._gid

    @property
    def name(self) -> str:
        return self._name

    @property
    def lights(self) -> list:
        return self._lights

    @property
    def all_on(self) -> bool:

        try:
            addr = self._bridge._url + f"/groups/{self._gid}"
            self._all_on = self._bridge.api_get(addr)["state"]["all_on"]

        except Exception:
            pass

        return self._all_on

    @all_on.setter
    def all_on(self, value: bool) -> None:
        for light in self._lights:
            light.on = value

    @property
    def brightness(self) -> int:
        try:
            addr = self._bridge._url + f"/groups/{self._gid}"
            self._brightness = self.value_to_percent(
                self.api_get(addr)["action"]["bri"], 255
            )

        except Exception:
            pass

        return self._brightness

    @brightness.setter
    def brightness(self, value: int) -> None:
        if value > 0 and value <= 100:
            self._brightness = value
            addr = self._bridge._url + f"/groups/{self._gid}/action"
            val = self.percent_to_value(self._brightness, 255)
            self.api_put(addr, f'{{"bri": {val}}}')



import py_cui.keys as Keys
import py_cui.grid as Grid
import py_cui.widgets as Widgets

from py_cui.debug import PyCUILogger as Logger
from huetui.backend.bridge import Bridge


class SceneMenu(Widgets.ScrollMenu):
    """Scene menu widget.

    Args:
        title (str): menu title
        grid (Grid): grid object
        row (int): x position of the menu
        column (int): y position of the menu
        row_span (int): x width of the menu
        column_span (int): y height of the menu
        padx (int): x padding
        pady (int): y padding
        logger (Logger): logger object
        bridge (Bridge): bridge object
    """

    def __init__(
        self,
        id,
        title: str,
        grid: Grid,
        row: int,
        column: int,
        row_span: int,
        column_span: int,
        padx: int,
        pady: int,
        logger: Logger,
        bridge: Bridge,
        master,
    ) -> None:
        self.bridge = bridge
        self.scenes = bridge.scenes
        self.groups = bridge.groups
        self.master = master
        super(SceneMenu, self).__init__(
            id, title, grid, row, column, row_span, column_span, padx, pady, logger
        )
        self.set_help_text(
            "Scenes: arrow keys to navigate,"
            " ENTER to select,"
            " ESC to exit"
        )
        self.add_item_list([scene.name for scene in self.scenes])
        self.add_key_command(Keys.KEY_ENTER, self.popup)

    def popup(self) -> None:
        """popup to select which group to set the scene to"""
        groups = [group.name for group in self.groups]
        self.master.show_menu_popup(
            "To which group?", [group.name for group in self.groups], self.set_scene
        )

    def set_scene(self, group: str) -> None:
        """set scene to a group. needs to be called from popup.

        Args:
            group (str): group name to set the scene to
        """
        scene = self.scenes[self.get_selected_item_index()]
        target = self.bridge.group_by_name(group)
        scene.set_to_group(target)



from typing import NamedTuple
from huetui.backend.utils import Util
from enum import Enum
import huetui.backend.bridge as Bridge
import colorsys


class RGB(NamedTuple):
    """Named tuple for RGB values.

    Args:
        NamedTuple (class): inherits from NamedTuple
    """

    r: int = 0
    g: int = 0
    b: int = 0


class HSV(NamedTuple):
    """Named tuple for HSV values.

    Args:
        NamedTuple (class): inherits from NamedTuple
    """

    h: int = 0
    s: int = 0
    v: int = 0


class Parameter(Enum):
    """
    Enum for the parameters of the light.
    """

    ON = "on"
    NAME = "name"
    BRIGHTNESS = "bri"
    HUE = "hue"
    SATURATION = "sat"
    REACHABLE = "reachable"


class Light(Util):
    """Class representing a light

    Args:
        Util (class): parent class containing some utility methods

    Attributes:
        _bridge (class): bridge the light is connected to
        _lid (int): id of the light
        _on (bool): on/off state of the light
        _name (str): name of the light
        _brightness (int): brightness of the light in percent
        _hue (int): hue of the light
        _saturation (int): saturation of the light
        _color (str): color of the light as rgb
        _reachable (bool): reachable state of the light
    """

    def __init__(self, bridge: Bridge, lid: int):
        """Contructor for Light class

        Args:
            bridge (class): bridge the light is connected to
            lid (int): id of the light
        """
        self._bridge = bridge

        self._lid = lid
        self._on = None
        self._name = None

        self._brightness = None
        self._hue = None
        self._saturation = None
        self._color = None

        self._reachable = None

        # try to get initial values from api
        try:
            self._on = self.get_api_param(Parameter.ON)
            self._brightness = self.get_api_param(Parameter.BRIGHTNESS)
            self._hue = self.get_api_param(Parameter.HUE)
            self._saturation = self.get_api_param(Parameter.SATURATION)
            self._reachable = self.get_api_param(Parameter.REACHABLE)

        except Exception:
            pass

    def __str__(self) -> str:
        return "<class 'Light'> lid: {}".format(self._lid)

    def __repr__(self) -> str:
        return "<class 'Light'> lid: {}".format(self._lid)

    def put_api_param(self, param: Parameter, value: str):
        """Update the light's state on the bridge

        Args:
            param (str): parameter to update
            value (str): value to update the parameter to

        Returns:
            bool: update successfull
        """
        try:

            data = f'{{"{param.value}":{value}}}'
            addr = self._bridge.url + f"/lights/{self._lid}/state"

            self.api_put(addr, data)

            return True

        except Exception as e:
            print(e)
            return False

    def get_api_param(self, param: Parameter):
        """Get the light's parameter from the api

        Args:
            param (str): parameter to get

        Returns:
            bool, str, int: value of the parameter
        """
        addr = self._bridge.url + f"/lights/{self._lid}"

        resp_dict = self.api_get(addr)
        val = resp_dict["state"][param.value]

        return val

    @staticmethod
    def hsv_to_rgb(h: int, s: int, v: int) -> RGB:
        """Converts hsv(65535, 100, 100) to rgb(255,255,255)

        Args:
            h (int): hue value
            s (int): saturation value
            v (int): brightness value

        Returns:
            RGB: red, green and blue, as a namedtuple
        """

        # normalize hue, saturation and brightness to 0-1
        h = h / 65535
        s = s / 100
        v = v / 100

        # convert to rgb
        (r, g, b) = colorsys.hsv_to_rgb(h, s, v)

        return RGB(round(r * 255), round(g * 255), round(b * 255))

    @staticmethod
    def rgb_to_hsv(r: int, g: int, b: int) -> HSV:
        """Converts rgb(255, 255, 255) to hsv(65535, 100, 100)

        Args:
            r (int): red value
            g (int): green value
            b (int): blue value

        Returns:
            HSV: hue, saturation and value, as a namedtuple
        """

        # normalize r, g, b to 0-1
        r = r / 255
        g = g / 255
        b = b / 255

        # convert to hsv
        (h, s, v) = colorsys.rgb_to_hsv(r, g, b)

        return HSV(round(h * 65535), round(s * 100), round(v * 100))

    @property
    def lid(self) -> int:
        return self._lid

    @property
    def on(self) -> bool:
        # try to get value from api
        try:
            self._on = self.get_api_param(Parameter.ON)
        except Exception:
            pass

        return self._on

    @on.setter
    def on(self, value: bool) -> None:
        self._on = value
        self.put_api_param(Parameter.ON, str(self._on).lower())

    @property
    def name(self) -> str:
        # try to get value from api
        try:
            addr = self._bridge.url + f"/lights/{self._lid}"

            resp_dict = self.api_get(addr)
            val = resp_dict[Parameter.NAME.value]
            self._name = val

        except Exception as e:
            pass

        return self._name

    @property
    def brightness(self) -> int:
        # try to get value from api
        try:
            # convert value to percent
            self._brightness = self.value_to_percent(
                self.get_api_param(Parameter.BRIGHTNESS), 255
            )
        except Exception:
            pass

        return self._brightness

    @brightness.setter
    def brightness(self, value: int) -> None:
        # convert percentage to value (0-255)
        if value > 0 and value <= 100 and self.on and self.reachable:
            self._brightness = value
            self.put_api_param(
                Parameter.BRIGHTNESS, self.percent_to_value(self._brightness, 255)
            )

    @property
    def hue(self) -> int:
        # try to get value from api
        try:
            self._hue = self.get_api_param(Parameter.HUE)
        except Exception:
            pass

        return self._hue

    @hue.setter
    def hue(self, value: int) -> None:
        self._hue = value
        self.put_api_param(Parameter.HUE, self._hue)

    @property
    def saturation(self) -> int:
        # try to get value from api
        try:
            # convert value to percent
            self._saturation = self.value_to_percent(
                self.get_api_param(Parameter.SATURATION), 255
            )
        except Exception:
            pass

        return self._saturation

    @saturation.setter
    def saturation(self, value: int) -> None:
        # convert percentage to value (0-255)
        self._saturation = value
        self.put_api_param(
            Parameter.SATURATION, self.percent_to_value(self._saturation, 255)
        )

    @property
    def color(self) -> RGB:
        # try to get most recent values from api
        try:
            self._hue = self.get_api_param(Parameter.HUE)
            self._saturation = self.value_to_percent(
                self.get_api_param(Parameter.SATURATION), 255
            )
            self._brightness = self.value_to_percent(
                self.get_api_param(Parameter.BRIGHTNESS), 255
            )

        except Exception:
            pass

        self._color = self.hsv_to_rgb(self._hue, self._saturation, self._brightness)

        return self._color

    @color.setter
    def color(self, rgb: RGB) -> None:
        # convert rgb to hsv
        hsv = self.rgb_to_hsv(rgb.r, rgb.g, rgb.b)

        self._color = rgb

        self._hue = hsv.h
        self._saturation = hsv.s
        self._brightness = hsv.v

        # update api
        self.put_api_param(Parameter.HUE, self._hue)
        self.put_api_param(
            Parameter.SATURATION, self.percent_to_value(self._saturation, 255)
        )
        self.put_api_param(
            Parameter.BRIGHTNESS, self.percent_to_value(self._brightness, 255)
        )

    @property
    def reachable(self) -> bool:
        # try to get value from api
        try:
            self._reachable = self.get_api_param(Parameter.REACHABLE)
        except Exception:
            pass

        return self._reachable

    def toggle(self) -> None:
        """toggle a light on or off"""
        self.on = not self.on



from huetui.backend.group import Group
from huetui.backend.utils import Util
import huetui.backend.bridge as Bridge

class Scene(Util):
    """Class representing a scene.

    Args:
        Util (class): parent class containing some utility methods
    """
    def __init__(self, bridge: Bridge, sid: str, name: str, lights: list) -> None:
        self._bridge = bridge
        self._sid = sid
        self._name = name
        self._lights = lights

    def __str__(self) -> str:
        return "<class 'Scene'> sid: {}".format(self._sid)

    def __repr__(self) -> str:
        return "<class 'Scene'> sid: {}".format(self._sid)

    @property
    def sid(self) -> str:
        return self._sid

    @property
    def name(self) -> str:
        return self._name

    @property
    def lights(self) -> list:
        return self._lights

    def set_to_group(self, group: Group) -> bool:
        """Set a group to the scene.

        Args:
            group (class): group to set the scene to
        """
        addr = self._bridge._url + f"/groups/{group.gid}/action"
        payload = f"{{\"scene\": \"{self._sid}\"}}"
        return self.api_put(addr, payload)



from dataclasses import dataclass
from huetui.backend.utils import Util
from huetui.backend.light import Light, RGB
from huetui.backend.scene import Scene
from huetui.backend.group import Group
from os.path import expanduser
from sys import exit as sysexit
from traceback import print_exc
from colorthief import ColorThief

import random


@dataclass
class Info:
    """
    Dataclass for the bridge's metadata.
    """

    name: str
    id: str
    mac: str
    ip: str
    sw_version: str
    api_version: str


class Bridge(Util):
    """Class representing a philips hue bridge.

    Args:
        Util (class): parent class containing some utility methods

    Attributes:
        _url (str): url of the bridge
        _lights (list): list of lights connected to the bridge
        _scenes (list): list of scenes defined on bridge
        _groups (list): list of groups defined on bridge
        _info (class): bridge's metadata
    """

    def __init__(self, url: str) -> None:
        self._url = url
        self._lights = []
        self._scenes = []
        self._groups = []

        self._info: Info = None

        # init bridge from api
        try:
            self._init_lights_from_api()
            self._init_scenes_from_api()
            self._init_groups_from_api()
            self._init_info_from_api()
        except Exception:
            print("Could not connect to bridge. Reason:")
            print_exc(limit=1)
            sysexit(1)

    def _init_lights_from_api(self) -> None:
        """
        initialize all lights by querying the api
        """
        addr = self._url + f"/lights"
        lights = self.api_get(addr)

        for lid in lights:
            self._lights.append(Light(self, int(lid)))

    def _init_scenes_from_api(self) -> None:
        """
        initialize all scenes by querying the api
        """
        addr = self._url + f"/scenes"
        scenes = self.api_get(addr)

        # for each scene
        for sid in scenes:
            name = scenes[sid]["name"]
            lights = []

            # find lights from scene in bridge
            for lid in scenes[sid]["lights"]:
                light = self.light_by_id(int(lid))

                # check if light is found
                if light:
                    lights.append(light)

            # create scene object and append
            self._scenes.append(Scene(self, sid, name, lights))

    def _init_groups_from_api(self) -> None:
        """
        initialize all groups by querying the api
        """
        addr = self._url + f"/groups"
        groups = self.api_get(addr)

        # for each scene
        for gid in groups:
            name = groups[gid]["name"]
            lights = []

            # find lights from scene in bridge
            for lid in groups[gid]["lights"]:
                light = self.light_by_id(int(lid))

                # check if light is found
                if light:
                    lights.append(light)

            # create scene object and append
            self._groups.append(Group(self, int(gid), name, lights))

    def _init_info_from_api(self) -> None:
        """
        initialize the bridge's metadata by querying the api
        """
        addr = self._url + f"/config"
        info = self.api_get(addr)

        self._info = Info(
            name=info["name"],
            id=info["bridgeid"],
            mac=info["mac"],
            ip=info["ipaddress"],
            sw_version=info["swversion"],
            api_version=info["apiversion"],
        )

    def light_by_id(self, lid: int) -> Light:
        """find light by id

        Args:
            lid (int): id of the light

        Returns:
            Light: light object with passed id
        """
        for light in self._lights:
            if light.lid == lid:
                return light
        return None

    def light_by_name(self, name: str) -> Light:
        """find light by name

        Args:
            name (str): name of the light

        Returns:
            Light: light object with passed name
        """
        for light in self._lights:
            if light.name == name:
                return light
        return None

    def scene_by_id(self, sid: str) -> Scene:
        """find scene by id

        Args:
            sid (str): id of the scene

        Returns:
            Scene: scene object with passed id
        """
        for scene in self._scenes:
            if scene.sid == sid:
                return scene
        return None

    def scene_by_name(self, name: str) -> Scene:
        """find scene by name

        Args:
            name (str): name of the scene

        Returns:
            Scene: scene object with passed name
        """
        for scene in self._scenes:
            if scene.name == name:
                return scene
        return None

    def group_by_id(self, gid: int) -> Group:
        """find group by id

        Args:
            gid (str): id of the group

        Returns:
            Group: group object with passed id
        """
        for group in self._groups:
            if group.gid == gid:
                return group
        return None

    def group_by_name(self, name: str) -> Group:
        """find group by name

        Args:
            name (str): name of the group

        Returns:
            Group: group object with passed name
        """
        for group in self._groups:
            if group.name == name:
                return group
        return None

    def scene_by_name(self, name: str) -> Scene:
        """find scene by name

        Args:
            name (str): name of the scene

        Returns:
            Scene: scene object with passed name
        """
        for scene in self._scenes:
            if scene.name == name:
                return scene
        return None

    def set_scene(self, group: Group, scene: Scene) -> bool:
        """set a scene for a group

        Args:
            group (Group): group object
            scene (Scene): scene object

        Returns:
            bool: True if successful, False otherwise
        """
        return scene.set_to_group(group)

    @property
    def url(self) -> str:
        return self._url

    @property
    def lights(self) -> list:
        return self._lights

    @property
    def scenes(self) -> list:
        return self._scenes

    @property
    def groups(self) -> list:
        return self._groups

    @property
    def info(self) -> Info:
        return self._info

    def info_as_str(self) -> str:
        """return bridge's metadata as string

        Returns:
            str: bridge's metadata as string
        """
        return f"Name:\t {self._info.name}\nID:\t {self._info.id}\nIP:\t {self._info.ip}\nMAC:\t {self._info.mac}\nSW:\t v{self._info.sw_version}\nAPI:\t v{self._info.api_version}"

    def set_lights_from_image(self, file) -> None:
        """set lights from image

        Args:
            file (str): path to image file
        """
        colors = ColorThief(expanduser(file)).get_palette(color_count=5, quality=50)
        rgbs = [RGB(*c) for c in colors]

        for light in self.lights:
            if light.on:
                light.color = random.choice(rgbs)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="huetui",
    version="2.1",
    author="channel-42",
    license="MIT",
    author_email="info@devls.de",
    description="A TUI for Philips Hue",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/channel-42/hue-tui",
    scripts=["bin/huetui"],
    packages=setuptools.find_packages(),
    python_requires='>=3.10',
    install_requires=[
        "urllib3",
        "click",
        "typing",
        "py-cui",
        "colorthief",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
    ],
)
